@@ -2,7 +2,7 @@
 # coding: UTF-8
 #
 # Author:   Dawid Laszuk
-# Contact:  laszukdawid@gmail.com
+# Contact:  https://github.com/laszukdawid/PyEMD/issues
 #
 # Feel free to contact for any information.
 """
@@ -11,25 +11,12 @@
 
 from __future__ import print_function
 
-import itertools
 import logging
 import numpy as np
 
+from collections import defaultdict
 from multiprocessing import Pool
-
-# Python3 handles mutliprocessing much better.
-# For Python2 we need to pickle instance differently.
-import sys
-if sys.version_info[0] < 3:
-    import copy_reg as copy_reg
-    import types
-    def _pickle_method(m):
-        if m.im_self is None:
-            return getattr, (m.im_class, m.im_func.func_name)
-        else:
-            return getattr, (m.im_self, m.im_func.func_name)
-
-    copy_reg.pickle(types.MethodType, _pickle_method)
+from typing import Dict, List, Optional, Sequence, Tuple, Union
 
 
 class EEMD:
@@ -57,6 +44,21 @@ class EEMD:
         One can pass EMD object defined outside, which will be
         used to compute IMF decompositions in each trial. If none
         is passed then EMD with default options is used.
+    parallel : bool (default: False)
+        Flag whether to use multiprocessing in EEMD execution.
+        Since each EMD(s+noise) is independent this should improve execution
+        speed considerably.
+        *Note* that it's disabled by default because it's the most common
+        problem when EEMD takes too long time to finish.
+        If you set the flag to True, make also sure to set `processes` to
+        some reasonable value.
+    processes : int or None (optional)
+        Number of processes harness when executing in parallel mode.
+        The value should be between 1 and max that depends on your hardware.
+    separate_trends : bool (default: False)
+        Flag whether to isolate trends from each EMD decosition into a separate component.
+        If `true`, the resulting EEMD will contain ensemble only from IMFs and
+        the mean residue will be stacked as the last element.
 
     References
     ----------
@@ -69,42 +71,40 @@ class EEMD:
 
     noise_kinds_all = ["normal", "uniform"]
 
-    def __init__(self, trials=100, noise_width=0.05, ext_EMD=None, parallel=True, **config):
+    def __init__(self, trials: int=100, noise_width: float=0.05, ext_EMD=None, parallel: bool=False, **kwargs):
 
         # Ensemble constants
         self.trials = trials
         self.noise_width = noise_width
+        self.separate_trends = bool(kwargs.get('separate_trends', False))
 
         self.random = np.random.RandomState()
-        self.noise_kind = "normal"
+        self.noise_kind = kwargs.get('noise_kind', 'normal')
         self.parallel = parallel
+        self.processes = kwargs.get('processes')  # Optional[int]
+        if self.processes is not None and not self.parallel:
+            self.logger.warning("Passed value for process has no effect when `parallel` is False.")
 
         if ext_EMD is None:
             from PyEMD import EMD
-            self.EMD = EMD()
+            self.EMD = EMD(**kwargs)
         else:
             self.EMD = ext_EMD
 
-        self.E_IMF = None
-        self.residue = None
+        self.E_IMF = None  # Optional[np.ndarray]
+        self.residue = None  # Optional[np.ndarray]
+        self._all_imfs = {}
 
-        # Update based on options
-        for key in config.keys():
-            if key in self.__dict__.keys() or key == "processes":
-                self.__dict__[key] = config[key]
-            elif key in self.EMD.__dict__.keys():
-                self.EMD.__dict__[key] = config[key]
-
-    def __call__(self, S, T=None, max_imf=-1):
+    def __call__(self, S: np.ndarray, T: Optional[np.ndarray]=None, max_imf: int=-1) -> np.ndarray:
         return self.eemd(S, T=T, max_imf=max_imf)
 
-    def __getstate__(self):
+    def __getstate__(self) -> Dict:
         self_dict = self.__dict__.copy()
         if 'pool' in self_dict:
             del self_dict['pool']
         return self_dict
 
-    def generate_noise(self, scale, size):
+    def generate_noise(self, scale: float, size: Union[int, Sequence[int]]) -> np.ndarray:
         """
         Generate noise with specified parameters.
         Currently supported distributions are:
@@ -135,11 +135,11 @@ class EEMD:
 
         return noise
 
-    def noise_seed(self, seed):
+    def noise_seed(self, seed: int) -> None:
         """Set seed for noise generation."""
         self.random.seed(seed)
 
-    def eemd(self, S, T=None, max_imf=-1):
+    def eemd(self, S: np.ndarray, T: Optional[np.ndarray]=None, max_imf: int=-1) -> np.ndarray:
         """
         Performs EEMD on provided signal.
 
@@ -150,7 +150,7 @@ class EEMD:
         ----------
         S : numpy array,
             Input signal on which EEMD is performed.
-        T : numpy array, (default: None)
+        T : numpy array or None, (default: None)
             If none passed samples are numerated.
         max_imf : int, (default: -1)
             Defines up to how many IMFs each decomposition should
@@ -176,8 +176,7 @@ class EEMD:
         # For trial number of iterations perform EMD on a signal
         # with added white noise
         if self.parallel:
-            processes = None if "processes" not in self.__dict__ else self.__dict__["processes"]
-            pool = Pool(processes=processes)
+            pool = Pool(processes=self.processes)
 
             all_IMFs = pool.map(self._trial_update, range(self.trials))
 
@@ -186,25 +185,48 @@ class EEMD:
         else:  # Not parallel
             all_IMFs = map(self._trial_update, range(self.trials))
 
-        all_IMFs_1, all_IMFs_2 = itertools.tee(all_IMFs, 2)
+        self._all_imfs = defaultdict(list)
+        for (imfs, trend) in all_IMFs:
 
-        max_imfNo = max([IMFs.shape[0] for IMFs in all_IMFs_1])
+            # A bit of explanation here.
+            # If the `trend` is not None, that means it was intentionally separated in the decomp process.
+            # This might due to `separate_trends` flag which means that trends are summed up separately
+            # and treated as the last component. Since `proto_eimfs` is a dict, that `-1` is treated literally
+            # and **not** as the *last position*. We can then use that `-1` to always add it as the last pos
+            # in the actual eIMF, which indicates the trend.
+            if trend is not None:
+                self._all_imfs[-1].append(trend)
 
-        self.E_IMF = np.zeros((max_imfNo, N))
-        for IMFs in all_IMFs_2:
-            self.E_IMF[:IMFs.shape[0]] += IMFs
+            for imf_num, imf in enumerate(imfs):
+                self._all_imfs[imf_num].append(imf)
 
-        self.E_IMF /= self.trials
+        # Convert defaultdict back to dict and explicitly rename `-1` position to be {the last value} for consistency.
+        self._all_imfs = dict(self._all_imfs)
+        if -1 in self._all_imfs:
+            self._all_imfs[len(self._all_imfs)] = self._all_imfs.pop(-1)
+
+        for imf_num in self._all_imfs.keys():
+            self._all_imfs[imf_num] =  np.array(self._all_imfs[imf_num])
+
+        self.E_IMF = self.ensemble_mean()
         self.residue = S - np.sum(self.E_IMF, axis=0)
 
         return self.E_IMF
 
-    def _trial_update(self, trial):
-        # Generate noise
-        noise = self.generate_noise(self._scale, self._N)
-        return self.emd(self._S+noise, self._T, self.max_imf)
+    def _trial_update(self, trial) -> Tuple[np.ndarray, Optional[np.ndarray]]:
+        """A single trial evaluation, i.e. EMD(signal + noise).
 
-    def emd(self, S, T, max_imf=-1):
+        *Note*: Although `trial` argument isn't used it's needed for the (multiprocessing) map method.
+        """
+        noise = self.generate_noise(self._scale, self._N)
+        imfs = self.emd(self._S+noise, self._T, self.max_imf)
+        trend = None
+        if self.separate_trends:
+            imfs, trend = self.EMD.get_imfs_and_trend()
+        
+        return (imfs, trend)
+
+    def emd(self, S: np.ndarray, T: np.ndarray, max_imf: int=-1) -> np.ndarray:
         """Vanilla EMD method.
 
         Provides emd evaluation from provided EMD class.
@@ -212,15 +234,36 @@ class EEMD:
         """
         return self.EMD.emd(S, T, max_imf)
 
-    def get_imfs_and_residue(self):
+    def get_imfs_and_residue(self) -> Tuple[np.ndarray, np.ndarray]:
         """
         Provides access to separated imfs and residue from recently analysed signal.
-        :return: (imfs, residue)
+
+        Returns
+        -------
+        (imfs, residue) : (np.ndarray, np.ndarray)
+            Tuple that contains all imfs and a residue (if any).
+
         """
         if self.E_IMF is None or self.residue is None:
             raise ValueError('No IMF found. Please, run EMD method or its variant first.')
-        else:
-            return self.E_IMF, self.residue
+        return self.E_IMF, self.residue
+    
+    @property
+    def all_imfs(self):
+        """A dictionary with all computed imfs per given order."""
+        return self._all_imfs
+    
+    def ensemble_count(self) -> List[int]:
+        """Count of imfs observed for given order, e.g. 1st proto-imf, in the whole ensemble."""
+        return [len(imfs) for imfs in self._all_imfs.values()]
+
+    def ensemble_mean(self) -> np.ndarray:
+        """Pointwise mean over computed ensemble. Same as the output of `eemd()` method."""
+        return np.array([imfs.mean(axis=0) for imfs in self._all_imfs.values()])
+
+    def ensemble_std(self) -> np.ndarray:
+        """Pointwise standard deviation over computed ensemble."""
+        return np.array([imfs.std(axis=0) for imfs in self._all_imfs.values()])
 
 ###################################################
 ## Beginning of program
@@ -228,7 +271,6 @@ class EEMD:
 if __name__ == "__main__":
 
     import pylab as plt
-    global E_imfNo
     E_imfNo = np.zeros(50, dtype=np.int)
 
     # Logging options
@@ -245,8 +287,7 @@ if __name__ == "__main__":
     S = 3*np.sin(4*T) + 4*np.cos(9*T) + np.sin(8.11*T+1.2)
 
     # Prepare and run EEMD
-    eemd = EEMD()
-    eemd.trials = 50
+    eemd = EEMD(trials=50)
     eemd.noise_seed(12345)
 
     E_IMFs = eemd.eemd(S, T, max_imf)
